#include "userbase.h"

#include "stdhtml.html.h"

#define PN__RAW -2
#define PN__END -1

enum dtrcn {
	PN_PRL, PN_PPRL, PN_TITLE, PN_AUTHOR, PN_DATE,
	PN_FOREACH, PN_END, PN_PRLA,
	PN_IF_SUBINFO, PN_ELSE_SUBINFO, PN_FI_SUBINFO,
	PN_NREPLIES, PN_TEXTSIZE,
	PN_BADBADBAD
};

static char *dtrc [] = {
	"PRL", "PPRL", "TITLE", "AUTHOR", "DATE",
	"FOREACH", "END", "PRLA",
	"IF_SUBINFO", "ELSE_SUBINFO", "FI_SUBINFO",
	"NREPLIES", "TEXTSIZE",
	NULL
};

//************************************************************************
//
//	Http Request Parser &&
//	x-www-form-urlencoded Parser
//
//************************************************************************

#define HTTP_AUTH_USER		"x"
#define HTTP_MAX_TAG		256
#define HTTP_MAX_REQUEST	128

class http_request
{
	void cgi_parse (char*, int);
	char *Content;
	bool Authenticate (char*);
   public:
	http_request (FILE*);
	char Request [HTTP_MAX_REQUEST];
	bool BadRequest, Authorized;
	int argc;
	char **argn, *argv (char*), *argv (int);
	~http_request ();
};

static const char CLen [] = "Content-length:";
static const char Auth [] = "Authorization:";

http_request::http_request (FILE *F)
{
	char tag [HTTP_MAX_TAG], *c;
	bool post;
	int i, ContentLength;

	post = Authorized = false;
	BadRequest = true;
	argc = ContentLength = 0;
	Content = NULL;
	argn = NULL;

	if (!fgets (tag, HTTP_MAX_TAG, F)) return;
	if (strncmp (tag, "GET ", 4) && !(post = !strncmp (tag, "POST ", 5)))
		return;

	i = (post) ? 5 : 4;
	if ((c = strchr (tag + i, ' '))) *c = 0;
	strcpy (Request, tag + i);

	for (;;) {
		if (!fgets (tag, HTTP_MAX_TAG, F)) return;
		if (tag [0] == '\r' || tag [0] == '\n') break;

		if (!strncasecmp (tag, CLen, sizeof CLen - 1))
			ContentLength = strtoul (tag+sizeof CLen - 1, NULL,10);
		else if (!strncasecmp (tag, Auth, sizeof Auth - 1))
			Authorized = Authenticate (tag + sizeof Auth);
	}

	if (!Authorized) return;

	if (post && ContentLength) {
		char *form = (char*) alloca (ContentLength + 1);
		if (fread (form, 1, ContentLength, F) != (size_t) ContentLength)
			return;
		form [ContentLength] = 0;
		cgi_parse (form, ContentLength);
	}

	BadRequest = false;
}

bool http_request::Authenticate (char *l)
{
	while (isspace (*l)) ++l;
	if (strncmp (l, "Basic ", 6)) return false;
	l += 6;

	char mk [100], *c;
	char basic_cookie [50];
	sprintf (basic_cookie, HTTP_AUTH_USER":%s", Globals.HTTPPassword);
	for (c = l + 6; *c > ' '; c++);
	*c = 0;
	return strcmp (Encode64 (basic_cookie, mk), l) == 0; 
}

static char hex2char (char h, char l)
{
	unsigned char ret;
	ret = 16 * ((h >= '0' && h <= '9') ? h - '0' :
	      ((h <= 'F') ? h - 'A' : h - 'a') + 10);
	ret += (l >= '0' && l <= '9') ? l - '0' :
	      ((l <= 'F') ? l - 'A' : l - 'a') + 10;

	// Voodoo Shit
	return ret;
}

void http_request::cgi_parse (char *b, int e)
{
	char *tmp = (char*) alloca (e);
	int i, j, n;
	int field [30];		// enough for this application

	n = i = j = 0;
	while (b [i]) {
		field [n++] = j;
		while (b[i] && b [i] != '=')
			tmp [j++] = b [i++];
		tmp [j++] = 0;
		if (!b[i++]) break;
		while (b [i] && b [i] != '&')
			switch (b [i]) {
			case '+':
				tmp [j++] = ' ';
				++i;
				break;
			case '%':
				tmp [j++] = hex2char (b [i + 1], b [i + 2]);
				i += 3;
				break;
			default:
				tmp [j++] = b [i++];
			}
		if (b [i] == '&') ++i;
		tmp [j++] = 0;
	}

	if ((argc = n)) {
		argn = new char* [n];
		Content = new char [j];
		memcpy (Content, tmp, j);
		for (i = 0; i < n; i++)
			argn [i] = &Content [field [i]];
	}
}

char *http_request::argv (char *n)
{
	int i;
	for (i = 0; i < argc; i++)
	if (!strcmp (argn [i], n))
		return argn [i] + strlen (argn [i]) + 1;
	return NULL;
}

char *http_request::argv (int i)
{
	return (argn [i]) ? argn [i] + strlen (argn [i]) + 1 : NULL;
}

http_request::~http_request ()
{
	if (Content) delete [] Content;
	if (argn) delete [] argn;
}

//***********************************************************************
//
//	Http Server Respond:
//		http_reply_data		: send static file
//		DRequest		: constructed HTML
//
//***********************************************************************

static void http_reply_data (FILE *WWW, char *d, int len, char *type)
{
	fputs ("HTTP/1.0 200 Ok\r\n", WWW);
	fprintf	(WWW, "Content-Length: %i\r\n", len);
	fprintf (WWW, "Content-Type: %s\r\n", type);
	fputs ("Connection: Close\r\n", WWW);
	fputs ("\r\n", WWW);
	fwrite (d, 1, len, WWW);
	fflush (WWW);
}

static void http_reply_data (FILE *WWW, char *s, char *type = NULL)
{
	http_reply_data (WWW, s, strlen (s), (type)?type:(char*)"text/html");
}

static void request_authentication (FILE *WWW)
{
	fputs ("HTTP/1.0 401 Unauthorized\r\nConnection: Close\r\n", WWW);
	fputs ("WWW-Authenticate: Basic realm=\"\"\r\n\r\n", WWW);
	fflush (WWW);
}

struct pstr
{
	char *str;
	bool release;
	pstr *next;
	pstr (char *s, bool b) : next (NULL) {
		str = ((release = b)) ? StrDup (s) : s;
	}
	~pstr () {
		if (next) delete next;
		if (release) delete [] str;
	}
};

class DRequest
{
	pstr *Start, *End;
	int Totlen;
   public:
	DRequest ()	: Start (NULL), End (NULL), Totlen (0) {}
	void append (char*, bool = false);
	void number (int);
	void mmaped (char*, int);
	void http_send (FILE*);
	~DRequest ()	{ if (Start) delete Start; }
};

void DRequest::append (char *s, bool d)
{
	pstr *P = new pstr (s, d);
	if (End) End->next = P;
	else Start = P;
	End = P;
	Totlen += strlen (s);
}

void DRequest::number (int i)
{
	char c [20];
	sprintf (c, "%i", i);
	append (c, true);
}

void DRequest::mmaped (char *d, int l)
{
	char *c = new char [l + 1];
	memcpy (c, d, l);
	c [l] = 0;
	append (c, true);
}

void DRequest::http_send (FILE *f)
{
	fputs ("HTTP/1.0 200 Ok\r\n", f);
	fprintf	(f, "Content-Length: %i\r\n", Totlen);
	fputs ("Content-type: text/html\r\nConnection: Close\r\n\r\n", f);

	for (pstr *p = Start; p; p = p->next)
		fputs (p->str, f);
	fflush (f);
}

//*********************************************************************
//
//	Do Requests -- construct HTML from compiled templates
//
//*********************************************************************

struct dtcmd
{
	int type;
	char *str;
} *dtReplies = NULL;

#define DTYPE(x, y) dtReplies[x].type == y
#define DNTYPE(x, y) dtReplies[x].type != y

static void append_prla (DRequest &R, char *PRL)
{
	char c [MAX_PRL];
	R.append (CPofPRL (PRL, c), true);
}

static void inval (DRequest &R, int t)
{
	if (t >= 0 && dtrc [t]) R.append (dtrc [t]);
	else R.number (t);
}

static void show_date (char *PRL, DRequest &R)
{
	postnet_get H (PRL, false);

	if (H.rep != -1) R.append (H.date, true);
	else R.append ("Not available");
}

static int pn_prl_subinfo (DRequest &R, char *PRL, int i)
{
	postnet_get H (PRL, false);

	if (H.rep == -1) {
		while (DNTYPE (i, PN__END) && DNTYPE (i, PN_ELSE_SUBINFO))
			++i;
		++i;
		for (;DNTYPE (i, PN__END) && DNTYPE (i, PN_FI_SUBINFO); i++)
		switch (dtReplies [i].type) {
		case PN__RAW:	R.append (dtReplies [i].str); break;
		case PN_PRL:	R.append (PRL); break;
		case PN_PRLA:	append_prla (R, PRL); break;
		default:	inval (R, dtReplies [i].type); break;
		}
	} else {
		++i;
		for (;DNTYPE (i, PN__END) && DNTYPE (i, PN_ELSE_SUBINFO); i++)
		switch (dtReplies [i].type) {
		case PN__RAW:		R.append (dtReplies [i].str); break;
		case PN_PRL:		R.append (PRL); break;
		case PN_PRLA:		append_prla (R, PRL); break;
		case PN_NREPLIES:	R.number (H.rep); break;
		case PN_TEXTSIZE:	R.number (H.len); break;
		case PN_DATE:		R.append (H.date, true); break;
		default:		inval (R, dtReplies [i].type); break;
		}
		while (DNTYPE (i, PN__END) && DNTYPE (i, PN_FI_SUBINFO))
			++i;
	}

	return i;
}

static int pn_prl_foreach (DRequest &R, postnet_download &P, int i)
{
	int j = i + 1, k;

	for (k = 0; k < P.replies; k++)
		for (i = j; DNTYPE (i, PN__END) && DNTYPE (i, PN_END); i++)
		switch (dtReplies [i].type) {
		case PN__RAW:		R.append (dtReplies [i].str); break;
		case PN_PRL:		R.append (P.nPRL (k)); break;
		case PN_PPRL:		R.append (P.ThisPRL); break;
		case PN_TITLE:		R.append (P.nTitle (k)); break;
		case PN_AUTHOR:		R.append (P.nAuthor (k)); break;
		case PN_DATE:		R.append (P.nDate (k)); break;
		case PN_PRLA:		append_prla (R, P.nPRL (k)); break;
		case PN_IF_SUBINFO:	i = pn_prl_subinfo (R, P.nPRL (k), i);
					break;
		default:		inval (R, dtReplies [i].type); break;
		}

	for (i = j; DNTYPE (i, PN__END) && DNTYPE (i, PN_END); i++);

	return i;
}

static void pn_download (char *PRL, FILE *WWW)
{
	DRequest R;

	if (!goodPRL (PRL)) {
		return;
	}

	postnet_download P (PRL);

	if (P.success == PD_OK)
		for (int i = 0; DNTYPE (i, PN__END); i++)
		switch (dtReplies [i].type) {
		case PN__RAW:	R.append (dtReplies [i].str); break;
		case PN_PPRL:	R.append (P.Parent); break;
		case PN_PRL:	R.append (PRL); break;
		case PN_PRLA:	append_prla (R, PRL); break;
		case PN_TITLE:	R.append (P.Title); break;
		case PN_AUTHOR:	R.append (P.Author); break;
		case PN_NREPLIES:	R.number (P.replies); break;
		case PN_DATE:	show_date (PRL, R); break;
	
		case PN_FOREACH:	i = pn_prl_foreach (R, P, i); break;
		default:	inval (R, dtReplies [i].type); break;
		}

	else R.append ("<html><body> Failed </body></html>\n");

	R.http_send (WWW);
}

static void pn_root (FILE *WWW)
{
	DRequest R;
	R.append (roothtml1);
	R.append (Globals.ThisAuthor);
	R.append (roothtml2);
	R.http_send (WWW);
}

static void pn_enterw (http_request &HTTP, FILE *WWW)
{
	if (HTTP.argc && goodPRL (HTTP.argv (0)))
		pn_download (HTTP.argv (0), WWW);
	else {
		DRequest R;
		R.append (badaddr);
		R.http_send (WWW);
	}
}

static char myposts_head [] =
"<html><head><title> My posts </title></head>"
"<body> (make this nice!) <ul> my posts ";
static char myposts_body [] =
"<li> [%i] <a href=\"%s/%x.%x\"> %s </a> at <i> %s </i> <b>%s</b></li>\n";
static char myposts_tail [] =
"</ul> </body> </html>\n";

static void pn_myposts (FILE *WWW)
{
	// kludge. Shall be rewritten....
	char tmp [200], dateb [40];
	DRequest R;
	post *p;

	R.append (myposts_head);
	for (int i = Globals.TopPostID; i >= 0; i--)
	if ((p = (post*) Globals.Posts.intFind (i))) {
		sprintf (tmp, myposts_body, i, Globals.ThisPoint,
			 p->Key, p->rkey, p->title,
			 network_date (p->date, dateb),
			 (p->secret == SECRET_UNPOSTED) ? "(unposted)"
			  : ((p->zombie) ? "zombie" : ""));
		R.append (tmp, true);
	}
	R.append (myposts_tail);

	R.http_send (WWW);
}

static void pn_reply (http_request &HTTP, FILE *WWW)
{
	// kludge. Shall be rewritten....
	char tmp [1024];
	post *p = NULL;
	DRequest R;

	if (HTTP.argv (INPUT_PRL) && HTTP.argv (INPUT_SUBJECT)
	&& HTTP.argv (INPUT_COMMENT) && HTTP.argv (INPUT_COMMENT) [0]) {
		char *par = (!strcmp (HTTP.argv (INPUT_PRL), NEWTHREAD)) ?
			NULL : HTTP.argv (INPUT_PRL);
		p = Submit (par, HTTP.argv (INPUT_SUBJECT),
		HTTP.argv (INPUT_COMMENT), strlen (HTTP.argv (INPUT_COMMENT)));
	}

	if (p) {
		sprintf (tmp, submitted, Globals.ThisPoint, p->Key, p->rkey,
			 Globals.ThisPoint, p->Key, p->rkey);
		R.append (tmp);
	} else R.append ("<html><body> FAILED </body></html>");

	R.http_send (WWW);	
}

static void pn_content (char *PRL, FILE *WWW)
{
	DRequest R;
	postnet_get D (PRL, true);

	if (D.content)
		http_reply_data (WWW, D.content, D.len, D.type);
	else /* write me */;
}

static void pn_alien (char *r, FILE *WWW)
{
	DRequest R;

	R.append ("<html><head><title> I don't know that "
		  "</title></head><body> I dont understand :<tt>");
	R.append (r);
	R.append ("</tt></body></html>");
	R.http_send (WWW);
}

static void pn_enterreply (char *PRL, FILE *WWW)
{
	// kludge. Shall be rewritten....
	char tmp [2048];
	sprintf (tmp, reply_page, PRL, PRL, PRL);
	DRequest R;
	R.append (tmp);
	R.http_send (WWW);
}
 
//*********************************************************************
//
//	The Http Server Kernel.
//
//*********************************************************************

static void pn_config (http_request&, FILE*);
static void loadtemplates ();

static void do_http_request (http_request &HTTP, FILE *Fo)
{
	if (isPRLaddr (HTTP.Request))
		pn_download (HTTP.Request, Fo);
	else if (!strcmp (RELOAD_TEMPLATES, HTTP.Request))
		loadtemplates ();
	else if (!strcmp ("/", HTTP.Request))
		pn_root (Fo);
	else if (!strcmp (ENTER_WRAPPER, HTTP.Request))
		pn_enterw (HTTP, Fo);
	else if (!strcmp (MYPOSTS, HTTP.Request))
		pn_myposts (Fo);
	else if (!strcmp (REPLY_PL, HTTP.Request))
		pn_reply (HTTP, Fo);
	else if (!strncmp (HTTP.Request, "/text/", 6))
		pn_content (HTTP.Request + 5, Fo);
	else if (!strncmp (HTTP.Request, "/reply/", 7))
		pn_enterreply (HTTP.Request + 6, Fo);
	else if (!strncmp (HTTP.Request, "/config", 7))
		pn_config (HTTP, Fo);
	else pn_alien (HTTP.Request, Fo);

	fclose (Fo);
}

static unsigned int RestrictedIP;
static bool freeport;

static void *http_interface (void*)
{
	int fd, cfd;

	if ((fd = listen_tcp (&Globals.HTTPPort)) == -1)
		if (freeport) {
			Globals.HTTPPort = 0;
			fd = listen_tcp (&Globals.HTTPPort);
		} else {
			printf ("Can't get TCP port %i.\n"
				"Wait till its free or use -i hc for"
				" a different port\n", Globals.HTTPPort);
			ShutDown ();
			return NULL;
		}

	printf ("HTTP Interface On http://localhost:%i\n", Globals.HTTPPort);

	while ((cfd = accept_client (fd)) != -1)
		if (peerIP (cfd) == RestrictedIP) {
			FILE *Fi = fdopen (cfd, "r");
			http_request HTTP (Fi);
			if (!HTTP.Authorized)
				request_authentication (fdopen (cfd, "w"));
			else if (!HTTP.BadRequest)
				do_http_request (HTTP, fdopen (cfd, "w"));
			fclose (Fi);
			close (cfd);
		} else close (cfd);

	return NULL;
}

void forkhttp (unsigned long int VIP, bool fp)
{
	pthread_t t;

	RestrictedIP = VIP;
	freeport = fp;
	loadtemplates ();
	pthread_create (&t, NULL, http_interface, NULL);
	pthread_detach (t);
}

/* HCC RESOLVE:
pthread_create()
	http_interface (Pv)
END */

//*********************************************************************
//
//	HTML Template files
//
//*********************************************************************

static const char TEMPLATE_CMD [] = "PNTEMPLATE_";
#define sizeofTC sizeof TEMPLATE_CMD - 1

class ptcc
{
	char *arr, **cmds;
	int len, s;
	char *StrPDup (int);
   public:
	ptcc (char*, int, char**);
	int tocmd ();
	int dttype ();
	dtcmd *compile ();
};

ptcc::ptcc (char *a, int l, char **c)
{
	arr = a;
	len = l;
	cmds = c;
}

char *ptcc::StrPDup (int t)
{
	char *r = new char [t - s + 1];
	strncpy (r, arr + s, t - s);
	r [t - s] = 0;
	return r;
}

int ptcc::tocmd ()
{
	int i;

	for (i = s;; i++) {
		while (i < len && arr [i] != TEMPLATE_CMD [0]) ++i;
		if (i + (int) sizeof TEMPLATE_CMD > len) return len;
		if (!strncmp (arr + i, TEMPLATE_CMD, sizeofTC))
			return i;
	}
}

int ptcc::dttype ()
{
	int i = 0;
	char c [50];

	for (s += sizeofTC; isalpha (arr [s]) || arr [s] == '_'; s++)
		c [i++] = arr [s];
	c [i] = 0;

	for (i = 0; cmds [i]; i++)
		if (!strcmp (c, cmds [i])) break;
	return i;
}

dtcmd *ptcc::compile ()
{
	int di = 0, e;
	dtcmd *r, dtmp [200];		// enough for this program

	s = 0;

	for (; di < 100;) {
		e = tocmd ();
		if (e > s) {
			dtmp [di].type = -2;
			dtmp [di++].str = StrPDup (e);
			s = e;
		}
		if (s == len) break;
		dtmp [di++].type = dttype ();
	}
	dtmp [di].type = -1;

	r = new dtcmd [di];
	for (di = 0; dtmp [di].type != -1; di++)
		r [di] = dtmp [di];
	r [di].type = -1;

	return r;
}

static void loadtemplates ()
{
	printf ("compiling html templates...\n");
	zeroCopy Z ("rep.tmpl");

	if (Z.success == ZC_OK) {
		ptcc P (Z.d, Z.len, dtrc);
		if (dtReplies) delete [] dtReplies;
		dtReplies = P.compile ();
	} else {
		ptcc P (default_template, sizeof default_template - 1, dtrc);
		if (dtReplies) delete [] dtReplies;
		dtReplies = P.compile ();
	}
}

//*********************************************************************
//
//	Forms -- A Hairy Kludge
//		this is a kludge because the "right thing" is
//		a big issue. i'm already in a parenthesis from the
//		initial project. in the next version...
//
//*********************************************************************

static void pn_config_form (FILE *WWW)
{
	DRequest R;
	char tmp [2048];

	R.append ("<html><head><title> Config </title></head><body>\n");

	//************************ HTTP pass **********************
	char form_httppass [] =
	"<HR><form method=post action=/config.httppass>\n"
	"Change your HTTP server password"
	"<input type=text name=pass size=15 maxlength=15> <br>"
	"</form><hr>\n\n";
	R.append (form_httppass);

	//************************** LOCp *************************
	char form_locpass [] =
	"<HR><form method=post action=/config.locpass>\n"
	"Change your LOC server password"
	"<input type=text name=pass size=15 maxlength=15> <br>"
	"</form><hr>\n\n";
	if (!strncasecmp (Globals.ThisPoint, "/loc/", 5))
		R.append (form_locpass);

	//************************ FINGER *************************
	char form_finger1 [] =
	"<HR><form method=post action=/config.finger>\n"
	"<B> User Information </B> (text of post 0.0)<br>\n"
	"<textarea name=finger rows=10 cols=50>";
	char form_finger2 [] =
	"</textarea><br>"
	"<input type=checkbox name=crtf value=on %s> "
	" Can others reply to post 0.0? <br>"
	"<input type=submit value=Apply>"
	"</form><hr>\n\n";
	R.append (form_finger1);
	{
		zeroCopy Z ("0.txt");
		if (Z.success == ZC_OK) R.mmaped (Z.d, Z.len);
	}
	sprintf (tmp, form_finger2, (Globals.CanReplyToFinger)
			 ? "CHECKED" : "");
	R.append (tmp, true);

	//************************** BAN:( ************************
	char form_ban [] =
	"<HR><form method=post action=/config.ban>\n"
	"<B>BAN</B><br> in the future we shall be able to see the "
	"list of banned posters and un-ban them. For now, just "
	"ban them spamming trolls. <BR> BAN THIS ADDRESS"
	"<input type=text name=ban size=50 maxlengt=70><br>"
	"<h3> address can be <tt> /dns/host.com/9123 </tt>"
	"or <tt> /loc/loc.com/9009/user </tt> or"
	"<tt> /loc/loc.com/9090 </tt> for all users of loc.com."
	" We can't ban DNS domains yet</h3>\n"
	"</form><hr>\n\n";
	R.append (form_ban);

	//************************** GURU *************************
	char form_guru [] =
	"<HR><form method=post action=/config.guru>\n"
	"<B>GURU</B> internal changes\n<br>"
	"Old value <input type=text name=old size=50 maxlength=80><br>"
	"New value <input type=text name=new size=50 maxlength=80><br>"
	"<input type=checkbox name=global value=off> Global changes<br>"
	"<input type=submit value=\"Do it (are you really sure?)\">"
	"</form><hr>\n\n";
	R.append (form_guru);

	//************************* SHUTDOWN **********************
	char form_shutdown [] =
	"<HR><form method=post action=/config.shutdown>\n"
	"<input type=hidden name=verify value=ok>"
	"<input type=submit value=\"Shut Down POSTNET Server\">"
	"</form><hr>\n\n";
	R.append (form_shutdown);

	//*********************************************************
	R.append ("</body></html>");

	R.http_send (WWW);
}

static bool pn_config_guru (http_request &HTTP)
{
	if (HTTP.argv ("old") && HTTP.argv ("new")
	&& HTTP.argv ("old") [0] == '/' && HTTP.argv ("new") [0] == '/') {
		if (HTTP.argv ("global"))
			Guru_pattern_cp (HTTP.argv ("old"), HTTP.argv ("new"));
		else
			Guru_explict_cp (HTTP.argv ("old"), HTTP.argv ("new"));
		return true;
	}
	return false;
}

static bool pn_config_ban (http_request &HTTP)
{
	if (HTTP.argv ("ban") && HTTP.argv ("ban") [0] == '/') {
		Ban (HTTP.argv ("ban"));
		return true;
	}
	return false;
}

static bool pn_config_locpass (http_request &HTTP)
{
	return (HTTP.argv ("pass")) ?
		 change_LOC_passwd (HTTP.argv ("pass")) : false;
}

static bool pn_config_httppass (http_request &HTTP)
{
	if (HTTP.argv ("pass")) {
		delete [] Globals.HTTPPassword;
		Globals.HTTPPassword = StrDup (HTTP.argv ("pass"));
	}
	return true;
}

static bool pn_config_finger (http_request &HTTP)
{
	if (HTTP.argv ("finger")) {

		Globals.CanReplyToFinger = (HTTP.argv ("crtf")); 
		FILE *f = fopen ("0.txt", "w");
		if (f) {
			fputs (HTTP.argv ("finger"), f);
			fclose (f);
			return true;
		}
	}
	return false;
}

static bool pn_config_shutdown (http_request &HTTP)
{
	if (HTTP.argv ("verify") && !strcmp (HTTP.argv ("verify"), "ok"))
		ShutDown ();
	return true;
}

static char html_success [] =
"<html><head><title> OK </title></head><body> "
"Ok <a href=/config> back </a> </body></html>";
static char html_failure [] =
"<html><head><title> FAILED </title></head><body> Action Failed"
"<a href=/config> back </a> </body></html>";

static void pn_config (http_request &HTTP, FILE *WWW)
{
	char *r = HTTP.Request + 7;	/* + "/config" */
	bool succ = false;

	if (r [0] == 0) {
		pn_config_form (WWW);
		return;
	}

	if (!strcmp (r, ".guru")) succ = pn_config_guru (HTTP);
	else if (!strcmp (r, ".ban")) succ = pn_config_ban (HTTP);
	else if (!strcmp (r, ".finger")) succ = pn_config_finger (HTTP);
	else if (!strcmp (r, ".locpass")) succ = pn_config_locpass (HTTP);
	else if (!strcmp (r, ".httppass")) succ = pn_config_httppass (HTTP);
	else if (!strcmp (r, ".shutdown")) succ = pn_config_shutdown (HTTP);

	http_reply_data (WWW, (succ) ? html_success : html_failure);
}
